package dev.pfaff.lazy;

import java.lang.invoke.*;
import java.util.function.Supplier;

import static java.lang.invoke.MethodHandles.constant;
import static java.lang.invoke.MethodHandles.insertArguments;
import static java.lang.invoke.MethodType.methodType;

public final class Lazy<T> implements Supplier<T> {
	private static final MethodHandle MH_lazyInitConstant;
	private static final MethodHandle MH_lazyInitConstantSynchronized;
	private static final MethodHandle MH_lazyInit;
	private static final MethodHandle MH_lazyInitSynchronized;
	private static final MethodType MT_lazyInit_wo_args;
	private static final MethodType MT_lazyInitSynchronized_wo_args;

	static {
		try {
			MH_lazyInitConstant = MethodHandles.lookup().findStatic(Lazy.class,
					"lazyInitConstant",
					methodType(Object.class, CallSite.class, Supplier.class));
			MH_lazyInitConstantSynchronized = MethodHandles.lookup().findStatic(Lazy.class,
					"lazyInitConstantSynchronized",
					methodType(Object.class, CallSite.class, Supplier.class, InitializationStatus.class));
			MH_lazyInit = MethodHandles.lookup().findStatic(Lazy.class,
					"lazyInit",
					methodType(Object.class, CallSite.class, Supplier.class, Object[].class));
			MH_lazyInitSynchronized = MethodHandles.lookup().findStatic(Lazy.class,
					"lazyInitSynchronized",
					methodType(Object.class,
							CallSite.class,
							Supplier.class,
							InitializationStatus.class,
							Object[].class));
			MT_lazyInit_wo_args = MH_lazyInit.type()
					.dropParameterTypes(MH_lazyInit.type().parameterCount() - 1, MH_lazyInit.type().parameterCount());
			MT_lazyInitSynchronized_wo_args = MH_lazyInitSynchronized.type()
					.dropParameterTypes(MH_lazyInitSynchronized.type().parameterCount() - 1,
							MH_lazyInitSynchronized.type().parameterCount());
		} catch (ReflectiveOperationException e) {
			throw new RuntimeException(e);
		}
	}

	private static <T> T get(MethodHandle getter) {
		try {
			//noinspection unchecked
			return (T) getter.invoke();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private final MethodHandle getter;

	private static <T> T lazyInitConstant(CallSite callSite, Supplier<? extends T> supplier) {
		var value = supplier.get();
		callSite.setTarget(constant(callSite.type().returnType(), value));
		return value;
	}

	private static <T> T lazyInitConstantSynchronized(CallSite callSite, Supplier<? extends T> supplier,
													  InitializationStatus status) {
		synchronized (callSite) {
			if (status.initialized) return get(callSite.getTarget());
			var value = lazyInitConstant(callSite, supplier);
			status.initialized = true;
			return value;
		}
	}

	private static <T> T lazyInit(CallSite callSite, Supplier<MethodHandle> supplier, Object... arguments) {
		var handle = supplier.get();
		callSite.setTarget(handle);
		try {
			return (T) handle.invokeWithArguments(arguments);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private static <T> T lazyInitSynchronized(CallSite callSite, Supplier<MethodHandle> supplier,
											  InitializationStatus status, Object... arguments) {
		synchronized (callSite) {
			if (status.initialized) return get(callSite.getTarget());
			var returnValue = Lazy.<T>lazyInit(callSite, supplier, arguments);
			status.initialized = true;
			return returnValue;
		}
	}

	public static <T> MethodHandle makeLazyConstantHandle(Class<T> type, Supplier<? extends T> supplier,
														  boolean isSynchronized) {
		var callSite = new MutableCallSite(methodType(type));
		callSite.setTarget(isSynchronized
				? insertArguments(
				MH_lazyInitConstantSynchronized.asType(MH_lazyInitConstantSynchronized.type().changeReturnType(type)),
				0, callSite, supplier, new InitializationStatus())
				: insertArguments(MH_lazyInitConstant.asType(MH_lazyInitConstant.type().changeReturnType(type)),
						0,
						callSite,
						supplier));
		return callSite.dynamicInvoker();
	}

	public static MethodHandle makeLazyMethodHandle(MethodType methodType, Supplier<MethodHandle> supplier,
													boolean isSynchronized) {
		var MH_lazyInit = isSynchronized ? MH_lazyInitSynchronized : Lazy.MH_lazyInit;
		MH_lazyInit = MH_lazyInit
				.asType((isSynchronized ? MT_lazyInitSynchronized_wo_args : MT_lazyInit_wo_args)
						.changeReturnType(methodType.returnType())
						.appendParameterTypes(methodType.parameterArray()));

		var callSite = new MutableCallSite(methodType);
		MH_lazyInit = isSynchronized
				? insertArguments(MH_lazyInit, 0, callSite, supplier, new InitializationStatus())
				: insertArguments(MH_lazyInit, 0, callSite, supplier);
		callSite.setTarget(MH_lazyInit);
		return callSite.dynamicInvoker();
	}

	public Lazy(Supplier<? extends T> supplier, boolean isSynchronized) {
		this.getter = makeLazyConstantHandle(Object.class, supplier, isSynchronized);
	}

	public Lazy(Supplier<T> supplier) {
		this(supplier, false);
	}

	@Override
	public T get() {
		return get(getter);
	}

	private static final class InitializationStatus {
		private boolean initialized;
	}
}
